#lang racket/base
(require "global.rkt"
         "nums.rkt"
         "pc.rkt"
         "mem.rkt"
         "reg.rkt"
         "assembly.rkt"
         "cpu.rkt"
         "bios.rkt")

;; The interactive command shell: 

;; shell: -> void
;; note: the function implements a read-command-eval-loop
(define (shell)
  (with-handlers ([exn? (lambda (exn)
                          (display (exn-message exn))
                          (newline))])
    (print-cell (get-pc))
    (printf " % ")  ; input prompt
    (let ([command (read)])
      (cond
       [(pair? command) (install (assemble command))]
       [(and (exact-integer? command) (value? command)) (install command)]
       [(symbol? command) ((find-command command))]
       [else (default-action)])))
  (shell))

;; install : value -> address
;; effect: store value in mem[PC], print as command, increment PC
(define (install value)
  (set-mem! (get-pc) value)
  (print-cell (get-pc))
  (newline)
  (get&inc-pc))

;; Some additional shell commands

;; step : -> void
(define (step)
  (let ((the-value (get-mem (get&inc-pc))))
    (printf "~n~s~n" (disassemble the-value))
    (execute (sign the-value) (decode the-value))
    (print-regs)))

;; reset! : -> void
;; effect: set all memory cells, registers, and the PC to 0
(define (reset!)
  (vector-fill! mem-array 0)
  (vector-fill! reg-array 0)
  (set-pc! 0))

;; show : -> void
;; effect: print contents of a range of memory cells, interactive
(define (show)
  (let* ((low (prompt-read "start: "))
	 (high (prompt-read "end: ")))
    (if (and
	  (exact-integer? low) (address? low)
	  (exact-integer? high) (address? high)
	  (<= low high))
	(let S ((low low))
	  (print-cell low) (newline)
	  (if (= low high) (void) (S (add1 low))))
	(begin
	  (printf "please enter two numbers, n and m, such that n <= m~n")
	  (show)))))

;; print-cell : address -> void
;; effect: print content of memory cell 
(define (print-cell addr)
  (let ([val (get-mem addr)])
    (printf "[~a]: ~a ~s"
      (format-number addr DIGITS-IN-ADD "")
      (format-number val  DIGITS        (if (negative? val) "-" "+"))
      (disassemble val))))

;; print-regs : -> void
;; effect: print contents of registers
(define (print-regs)
  (let ((reg (box 0)))
    (printf "PC:~s" (get-pc))
    (for-each (lambda (val)
		(printf " R~s:~s" (unbox reg) val)
		(set-box! reg (add1 (unbox reg))))
      (vector->list reg-array))
    (printf "~n")))

;; Input and output auxiliaries

;; format-number : num num str -> str
(define (format-number n width prefix)
  (let* ((n-as-string (number->string (abs n)))
	 (spaces-of-n (string-length n-as-string))
	 (no-of-lders (- width spaces-of-n)))
    (string-append
      prefix
      (make-string no-of-lders #\0)
      n-as-string)))

;; prompt-read : str -> TST
;; effect: print string and return next S-expression read from standard input
(define (prompt-read str)
  (printf str)
  (read))

;; default-action : -> void
;; effect: print help
(define (default-action)
  (for-each (lambda (e) (printf "~a ~n" (entry-description e))) the-command-list)
  (printf "number :: set current cell to value and move to next cell~n"))

;; The table of shell commands

;; structure definition for command table:
(define-struct entry (name description action))

;; list of available shell commands, names, and comments
;; the-command-list : (listof (structure:entry sym str (-> void)))
(define the-command-list
  (list 
    (make-entry 'q "q :: quit" exit)
    (make-entry 'r "r :: reset" 
      (lambda () (when (eq? (prompt-read "reset? [yn] ") 'y) (reset!))))
    (make-entry 'x "x :: execute" 
      (lambda ()
	(when (eq? (prompt-read (format "execute starting at ~s? [yn] " (get-pc))) 'y)
	  (printf "executing ... ~n")
	  (run)
	  (printf "~n... done. ~n"))))
    (make-entry 'n "n :: next (step)" step)
    (make-entry 'm "m :: show memory" show)
    (make-entry 's "s :: show registers and pc" print-regs)
    (make-entry 'p "p :: set pc"
      (lambda ()
	(let ((new-pc (prompt-read "  new pc: ")))
	  (when (and (number? new-pc) (address? new-pc)) (set-pc! new-pc)))))
    (make-entry '+ "+ :: forward" 
      (lambda () 
	(let ((new-pc (add1 (get-pc))))
	  (when (address? new-pc) (set-pc! new-pc)))))
    (make-entry '- "- :: backward" 
      (lambda () 
	(let ((new-pc (sub1 (get-pc))))
	  (when (address? new-pc) (set-pc! new-pc)))))))

;; find-command : sym -> (-> void)
(define (find-command command)
  (let ((r (findf (lambda (x) (eq? (entry-name x) command)) the-command-list)))
    (if r (entry-action r) default-action)))

;; Start the shell:
(shell)
